import math
from datetime import timedelta
from typing import Any
from typing import Dict

from service_capacity_modeling.models.org.netflix.iso_date_math import (
    _iso_to_proto_duration,
)
from service_capacity_modeling.models.org.netflix.iso_date_math import _iso_to_timedelta
from service_capacity_modeling.models.org.netflix.iso_date_math import iso_to_seconds

DURATION_1H = timedelta(hours=1)
DURATION_4H = timedelta(hours=4)
DURATION_6H = timedelta(hours=6)
DURATION_1D = timedelta(days=1)
DURATION_7D = timedelta(days=7)
DURATION_14D = timedelta(days=14)
DURATION_1M = timedelta(days=30)
DURATION_3M = DURATION_1M * 3
DURATION_6M = DURATION_1M * 6
DURATION_9M = DURATION_1M * 9
DURATION_1Y = timedelta(days=365)
_4MiB = 4 * 1024 * 1024


class TimeSeriesConfiguration:
    def __init__(self, extra_model_arguments: Dict[str, Any]):
        self.read_interval_seconds = self.__get_read_interval_seconds(
            extra_model_arguments
        )
        self.retention_duration = str(
            extra_model_arguments.get("ts.hot.retention-interval", "unlimited")
        )
        self.seconds_per_slice = int(
            self.__get_slice_interval(self.retention_duration).total_seconds()
        )
        self.seconds_per_interval = self.__get_time_interval(
            self.seconds_per_slice, self.read_interval_seconds
        )
        self.buckets_per_id = self.__get_buckets_per_id(
            extra_model_arguments, self.seconds_per_interval
        )
        self.seconds_per_bucket = self.__get_seconds_per_bucket(
            self.read_interval_seconds, self.buckets_per_id, self.seconds_per_interval
        )
        self.coalesce_duration = self.__get_coalesce_duration(extra_model_arguments)
        self.consistency_target = self.__get_consistency_target(extra_model_arguments)
        self.accept_limit = self.__get_accept_limit(
            extra_model_arguments, self.retention_duration
        )
        self.read_amplification = self.__get_read_amplification(
            self.read_interval_seconds, self.seconds_per_interval, self.buckets_per_id
        )
        self.search_enabled = bool(extra_model_arguments.get("search.enabled"))

    # We force dependent parameters to be passed into these methods
    # instead of using self to make dependencies between them more explicit
    @staticmethod
    def __get_accept_limit(
        extra_model_arguments: Dict[str, Any], retention_duration: str
    ) -> str:
        if "ts.accept-limit" in extra_model_arguments:
            return _iso_to_proto_duration(extra_model_arguments["ts.accept-limit"])
        elif retention_duration == "unlimited":
            # For "all-time" tables we have to let them read and write anywhere
            return "0s"
        else:
            return _iso_to_proto_duration("PT10M")

    @staticmethod
    def __get_buckets_per_id(
        extra_model_arguments: Dict[str, Any], seconds_per_interval: int
    ) -> int:
        events_per_day_per_ts = int(
            extra_model_arguments.get("ts.events-per-day-per-ts", "10")
        )
        event_size = int(extra_model_arguments.get("ts.event-size", "1024"))
        bytes_per_interval = (
            seconds_per_interval * events_per_day_per_ts * event_size
        ) / 86400
        return round(max(min(8, bytes_per_interval / _4MiB), 1))

    @staticmethod
    def __get_time_interval(seconds_per_slice: int, read_interval_seconds: int) -> int:
        if seconds_per_slice != 0 and read_interval_seconds > seconds_per_slice:
            # default bucket interval to slice
            # interval to avoid compaction issue seen in C*
            return seconds_per_slice
        return read_interval_seconds

    @staticmethod
    def __get_seconds_per_bucket(
        read_interval_seconds: int, buckets: int, seconds_per_interval: int
    ) -> int:
        if buckets == 1:
            return seconds_per_interval
        # For smaller read intervals, default to 60. Else rotate buckets every 5 secs
        if read_interval_seconds <= 60:
            return 60
        return 5

    @staticmethod
    def __get_slice_interval(retention: str) -> timedelta:
        """
        Choice of slice interval tries to balance the following:
        - Minimal disk waste - the closer the slice interval is to the retention time,
        the more waste there will be
        - Minimal read amplification - when read interval is larger than slice interval,
        there will be a read for each slice within that interval
        - Reasonably frequent slice creations - if slices are created too infrequently,
        it becomes very slow to update slice configuration
        - Intuitive Slice Sizes - we should create slices at regular intervals to
        improve debugging
        """
        # Disable too many returns - the alternate
        # version of this with <6 returns is more complex
        # pylint: disable=too-many-return-statements
        if retention == "unlimited":
            return timedelta(seconds=0)
        retention_seconds = _iso_to_timedelta(retention)
        if retention_seconds < DURATION_6H:
            raise ValueError("Data retention should be at least 6 hours")
        # 6h - 1d retention - 4h slice
        if retention_seconds <= DURATION_1D:
            return DURATION_4H
        # 1d - 7d retention - 1d slice
        if retention_seconds <= DURATION_7D:
            return DURATION_1D
        # 7d - 14d retention - 7d slice
        if retention_seconds <= DURATION_14D:
            return DURATION_7D
        # 14d - 1m retention - 14d slice
        if retention_seconds <= DURATION_1M:
            return DURATION_14D
        # 1m - 3m retention - 1m slice
        if retention_seconds <= DURATION_3M:
            return DURATION_1M
        # 3m - 6m retention - 3m slice
        if retention_seconds <= DURATION_6M:
            return DURATION_3M
        # 6m - 1y retention - 6m slice
        if retention_seconds <= DURATION_1Y:
            return DURATION_6M
        # by default, return 1y slice
        return DURATION_1Y

    @staticmethod
    def __get_read_interval_seconds(extra_model_arguments: Dict[str, Any]) -> int:
        # default read duration of 1D
        read_interval = extra_model_arguments.get("ts.read-interval", "PT24H")
        if read_interval == "unlimited":
            return int(DURATION_1Y.total_seconds() * 100)
        return iso_to_seconds(read_interval)

    @staticmethod
    def __get_coalesce_duration(extra_model_arguments: Dict[str, Any]) -> str:
        # arbitrary defaults. Can be tuned later.
        fire_forget = extra_model_arguments.get("ts.fire-and-forget", False)
        if fire_forget:
            return "1s"
        return "0.001s"

    @staticmethod
    def __get_consistency_target(extra_model_arguments: Dict[str, Any]) -> str:
        val = extra_model_arguments.get("ts.read-consistency", "eventual").lower()
        if val in ("best-effort", "eventual"):
            return val
        else:
            # Default to read-your-writes consistency
            return "read-your-writes"

    @staticmethod
    def __get_read_amplification(
        read_interval_seconds: int, seconds_per_interval: int, buckets_per_id: int
    ) -> int:
        return math.ceil(read_interval_seconds / seconds_per_interval) * buckets_per_id



from pytest import approx

from service_capacity_modeling.capacity_planner import planner
from service_capacity_modeling.hardware import shapes
from service_capacity_modeling.interface import Buffer
from service_capacity_modeling.interface import BufferComponent
from service_capacity_modeling.interface import Buffers
from service_capacity_modeling.models.common import buffer_for_components
from service_capacity_modeling.models.common import cpu_headroom_target


def test_m5_same_headroom_as_r5():
    m5xl = planner.instance("m5.xlarge")
    r5xl = planner.instance("r5.xlarge")

    m5_headroom = cpu_headroom_target(m5xl)
    assert m5_headroom is not None
    r5_headroom = cpu_headroom_target(r5xl)
    assert r5_headroom is not None
    assert m5_headroom == r5_headroom


def test_larger_instance_needs_less_headroom():
    m5xl = cpu_headroom_target(planner.instance("m5.xlarge"))
    assert m5xl is not None
    m52xl = cpu_headroom_target(planner.instance("m5.2xlarge"))
    assert m52xl is not None
    assert m52xl < m5xl


def test_seventh_gen_instance_reflects_non_ht_boost():
    m6i4xl = cpu_headroom_target(planner.instance("m6i.4xlarge"))
    m7a4xl = cpu_headroom_target(planner.instance("m7a.4xlarge"))
    assert m7a4xl < m6i4xl
    assert m7a4xl == approx(0.16, rel=0.05)


def test_has_some_headroom():
    # For the purpose of headroom shapes and planner should both see the same
    # thing
    m5_instance = shapes.instance("m5.xlarge")
    r5_instance = planner.instance("r5.xlarge")
    assert m5_instance.cpu == 4
    m5_headroom = cpu_headroom_target(m5_instance)
    r5_headroom = cpu_headroom_target(r5_instance)
    assert m5_headroom is not None
    assert r5_headroom is not None
    assert m5_headroom > 0
    assert m5_headroom == r5_headroom

    assert 0.3 < r5_headroom < 0.5


def test_2xl_headroom_with_buffer():
    # TODO (jolynch) I think these numbers are probably just a little too
    #  aggressive with small instances, but maybe that's right ... let's see
    m6i2xl = planner.instance("m6i.2xlarge")

    headroom = cpu_headroom_target(m6i2xl)
    assert headroom == approx(0.28, rel=0.05)

    buffer_1_5x = Buffers(desired={BufferComponent.compute: Buffer(ratio=1.5)})
    headroom_with_1_5x_buffer = cpu_headroom_target(m6i2xl, buffers=buffer_1_5x)
    assert headroom_with_1_5x_buffer > headroom
    assert headroom_with_1_5x_buffer == approx(0.52, rel=0.05)

    buffer_2_0x = Buffers(desired={BufferComponent.compute: Buffer(ratio=2.0)})
    headroom_with_2_0x_buffer = cpu_headroom_target(m6i2xl, buffers=buffer_2_0x)
    assert headroom_with_2_0x_buffer > headroom_with_1_5x_buffer > headroom
    assert headroom_with_2_0x_buffer == approx(0.64, rel=0.05)


def test_2xl_headroom_with_buffer_no_hyperthreading():
    buffer_2x = Buffers(desired={BufferComponent.compute: Buffer(ratio=2.0)})
    m7a_2xl = planner.instance("m7a.2xlarge")
    effective_headroom = cpu_headroom_target(m7a_2xl, buffers=buffer_2x)

    # Seventh gen cpus are cores not threads, should be able to run with < 2x buffer

    assert cpu_headroom_target(m7a_2xl) == approx(0.22, rel=0.05)
    assert cpu_headroom_target(m7a_2xl) < effective_headroom


def test_default_buffer():
    buffers = Buffers(default=Buffer(ratio=4, sources={"default": Buffer(ratio=4)}))
    for component in BufferComponent:
        result = buffer_for_components(buffers, [component])
        assert result.ratio == 4
        assert component in result.components
        assert not result.sources


def test_common_buffer_fallbacks():
    buffers = Buffers(
        desired={
            "compute": Buffer(ratio=2.0, components=[BufferComponent.compute]),
            "storage": Buffer(ratio=3.2, components=[BufferComponent.storage]),
        }
    )

    assert buffer_for_components(buffers, [BufferComponent.cpu]) == Buffer(
        ratio=2.0,
        components=["compute", "cpu"],
        sources={"compute": buffers.desired["compute"]},
    )

    assert buffer_for_components(buffers, [BufferComponent.network]) == Buffer(
        ratio=2.0,
        components=["compute", "network"],
        sources={"compute": buffers.desired["compute"]},
    )

    assert buffer_for_components(buffers, [BufferComponent.disk]) == Buffer(
        ratio=3.2,
        components=["disk", "storage"],
        sources={"storage": buffers.desired["storage"]},
    )

    assert buffer_for_components(buffers, [BufferComponent.memory]) == Buffer(
        ratio=3.2,
        components=["memory", "storage"],
        sources={"storage": buffers.desired["storage"]},
    )


def test_precise_buffers():
    buffers = Buffers(
        desired={
            "custom-cpu": Buffer(ratio=2.0, components=[BufferComponent.cpu]),
            "custom-network": Buffer(ratio=2.5, components=[BufferComponent.network]),
            "custom-disk": Buffer(ratio=3.0, components=[BufferComponent.disk]),
            "custom-memory": Buffer(ratio=3.5, components=[BufferComponent.memory]),
        }
    )

    assert buffer_for_components(buffers, [BufferComponent.cpu]) == Buffer(
        ratio=2.0,
        components=["compute", "cpu"],
        sources={"custom-cpu": buffers.desired["custom-cpu"]},
    )
    assert buffer_for_components(buffers, [BufferComponent.network]) == Buffer(
        ratio=2.5,
        components=["compute", "network"],
        sources={"custom-network": buffers.desired["custom-network"]},
    )
    assert buffer_for_components(buffers, [BufferComponent.disk]) == Buffer(
        ratio=3.0,
        components=["disk", "storage"],
        sources={"custom-disk": buffers.desired["custom-disk"]},
    )
    assert buffer_for_components(buffers, [BufferComponent.memory]) == Buffer(
        ratio=3.5,
        components=["memory", "storage"],
        sources={"custom-memory": buffers.desired["custom-memory"]},
    )



import logging
import math
from enum import Enum
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import GIB_IN_BYTES
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import MEGABIT_IN_BYTES
from service_capacity_modeling.interface import MIB_IN_BYTES
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import compute_stateful_zone
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import sqrt_staffed_cores
from service_capacity_modeling.models.org.netflix.iso_date_math import iso_to_seconds

logger = logging.getLogger(__name__)


class ClusterType(str, Enum):
    strong = "strong"
    ha = "high-availability"


def _estimate_kafka_requirement(
    instance: Instance,
    desires: CapacityDesires,
    copies_per_region: int,
    hot_retention_seconds: float,
    zones_per_region: int = 3,
) -> Tuple[CapacityRequirement, Tuple[str, ...]]:
    """Estimate the capacity required for one zone given a regional desire

    The input desires should be the **regional** desire, and this function will
    return the zonal capacity requirement
    """
    normalized_to_mib = desires.model_copy(deep=True)
    read_mib_per_second = (
        normalized_to_mib.query_pattern.estimated_mean_read_size_bytes.mid
        / MIB_IN_BYTES
    )
    write_mib_per_second = (
        normalized_to_mib.query_pattern.estimated_mean_write_size_bytes.mid
        / MIB_IN_BYTES
    )

    # 1 concurrent reader, cpu time is per MiB read
    # 0.5 MiB / second
    rps = normalized_to_mib.query_pattern.estimated_read_per_second
    wps = normalized_to_mib.query_pattern.estimated_write_per_second
    normalized_to_mib.query_pattern.estimated_read_per_second = rps.scale(
        read_mib_per_second
    )
    normalized_to_mib.query_pattern.estimated_write_per_second = wps.scale(
        write_mib_per_second
    )
    # TODO: maybe revisit this?
    needed_cores = normalize_cores(
        core_count=sqrt_staffed_cores(normalized_to_mib),
        target_shape=instance,
        reference_shape=desires.reference_shape,
    )

    # (Nick): Keep 40% of available bandwidth for node recovery
    # (Joey): For kafka BW = BW_write + BW_reads
    #   let X = input write BW
    #   BW_in = X * RF
    #   BW_out = X * (consumers) + X * (RF - 1)
    bw_in = (
        (write_mib_per_second * MIB_IN_BYTES) * copies_per_region
    ) / MEGABIT_IN_BYTES
    bw_out = (
        (
            (read_mib_per_second * MIB_IN_BYTES)
            + ((write_mib_per_second * MIB_IN_BYTES) * (copies_per_region - 1))
        )
    ) / MEGABIT_IN_BYTES
    #   BW = (in + out) because duplex then 40% headroom.
    needed_network_mbps = max(bw_in, bw_out) * 1.40

    needed_disk = math.ceil(
        desires.data_shape.estimated_state_size_gib.mid * copies_per_region,
    )

    # Keep the last N seconds hot in cache
    needed_memory = (write_mib_per_second * hot_retention_seconds) // 1024

    # Now convert to per zone
    needed_cores = max(1, needed_cores // zones_per_region)
    needed_disk = max(1, needed_disk // zones_per_region)
    needed_memory = max(1, int(needed_memory // zones_per_region))
    needed_network_mbps = max(1, int(needed_network_mbps // zones_per_region))
    logger.debug(
        "Need (cpu, mem, disk) = (%s, %s, %s)",
        needed_cores,
        needed_memory,
        needed_disk,
    )

    return (
        CapacityRequirement(
            requirement_type="kafka-zonal",
            reference_shape=desires.reference_shape,
            cpu_cores=certain_int(needed_cores),
            mem_gib=certain_float(needed_memory),
            disk_gib=certain_float(needed_disk),
            network_mbps=certain_float(needed_network_mbps),
            context={
                "bw_in_mbps": bw_in,
                "bw_out_mbps": bw_out,
                "hot_retention_seconds": hot_retention_seconds,
                "replication_factor": copies_per_region,
            },
        ),
        ("spend", "disk", "mem"),
    )


def _upsert_params(cluster, params):
    if cluster.cluster_params:
        cluster.cluster_params.update(params)
    else:
        cluster.cluster_params = params


def _kafka_read_io(rps, io_size_kib, size_gib, recovery_seconds: int) -> float:
    # Get enough disk read IO capacity for some reads
    # In practice we have cache reducing this by 99% or more
    read_ios = rps * 0.05
    # Recover the node in 60 minutes, to do that we need
    size_kib = size_gib * (1024 * 1024)
    recovery_ios = max(1, size_kib / io_size_kib) / recovery_seconds
    # Leave 50% headroom for read IOs since generally we will hit cache
    return (read_ios + int(round(recovery_ios))) * 1.5


# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
def _estimate_kafka_cluster_zonal(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    desires: CapacityDesires,
    hot_retention_seconds,
    zones_per_region: int = 3,
    copies_per_region: int = 2,
    require_local_disks: bool = False,
    require_attached_disks: bool = False,
    required_zone_size: Optional[int] = None,
    max_regional_size: int = 150,
    max_local_disk_gib: int = 1024 * 5,
    min_instance_cpu: int = 2,
    min_instance_memory_gib: int = 12,
) -> Optional[CapacityPlan]:

    # Kafka doesn't like to deploy on single CPU instances or with < 12 GiB of ram
    if instance.cpu < min_instance_cpu or instance.ram_gib < min_instance_memory_gib:
        return None

    # if we're not allowed to use attached disks, skip EBS only types
    if instance.drive is None and require_local_disks:
        return None

    # if we're not allowed to use local disks, skip ephems
    if instance.drive is not None and require_attached_disks:
        return None

    # Kafka only deploys on gp3 drives right now
    if instance.drive is None and drive.name != "gp3":
        return None

    requirement, regrets = _estimate_kafka_requirement(
        instance=instance,
        desires=desires,
        zones_per_region=zones_per_region,
        copies_per_region=copies_per_region,
        hot_retention_seconds=hot_retention_seconds,
    )

    # Account for sidecars and base system memory
    base_mem = (
        desires.data_shape.reserved_instance_app_mem_gib
        + desires.data_shape.reserved_instance_system_mem_gib
    )

    # Kafka clusters in prod (tier 0+1) need at least 2 nodes per zone
    min_count = 1
    if desires.service_tier < 2:
        min_count = 2

    # Kafka read io / second is zonal
    normalized_to_mib = desires.model_copy(deep=True)
    read_mib_per_second: int = (
        int(normalized_to_mib.query_pattern.estimated_mean_read_size_bytes.mid)
        // MIB_IN_BYTES
        // zones_per_region
    )
    write_mib_per_second: int = (
        int(normalized_to_mib.query_pattern.estimated_mean_write_size_bytes.mid)
        // MIB_IN_BYTES
        // zones_per_region
    )

    # All Kafka IOs are sequential, so they can use the group size
    read_ios_per_second = max(1, (read_mib_per_second * 1024) // drive.seq_io_size_kib)
    write_ios_per_second = max(
        1, (write_mib_per_second * 1024) // drive.seq_io_size_kib
    )

    cluster = compute_stateful_zone(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=int(requirement.disk_gib.mid),
        needed_memory_gib=int(requirement.mem_gib.mid),
        needed_network_mbps=requirement.network_mbps.mid,
        # Leave 2x overhead for both reads and writes.
        # Readers are sequential, 100MiB/s read is 100Mib / block_size ios
        # Writers are sequential, 100MiB/s write is 100MiB / block_size ios
        required_disk_ios=lambda size, count: (
            _kafka_read_io(
                rps=read_ios_per_second / count,
                # Kafka does sequential IO
                io_size_kib=drive.seq_io_size_kib,
                size_gib=size,
                # Enough IO to recover a node in 60 minutes
                recovery_seconds=60 * 60,
            ),
            # Leave 100% IO headroom for writes
            copies_per_region * (write_ios_per_second / count) * 2,
        ),
        # Kafka can run up to 60% full on disk, let's stay safe at 40%
        required_disk_space=lambda x: x * 2.5,
        max_local_disk_gib=max_local_disk_gib,
        cluster_size=lambda x: x,
        min_count=max(min_count, 1),
        # Sidecars and Variable OS Memory
        # Kafka currently uses 8GiB fixed, might want to change to min(30, x // 2)
        reserve_memory=lambda instance_mem_gib: base_mem + 8,
    )

    # Communicate to the actual provision that if we want reduced RF
    params = {"kafka.copies": copies_per_region}
    _upsert_params(cluster, params)

    # Sometimes we don't want to modify cluster topology, so only allow
    # topologies that match the desired zone size
    if required_zone_size is not None and cluster.count != required_zone_size:
        return None

    # Kafka clusters generally should try to stay under some total number
    # of nodes. Orgs do this for all kinds of reasons such as
    #   * Security group limits. Since you must have < 500 rules if you're
    #       ingressing public ips)
    #   * Maintenance. If your restart script does one node at a time you want
    #       smaller clusters so your restarts don't take months.
    #   * NxN network issues. Sometimes smaller clusters of bigger nodes
    #       are better for network propagation
    if cluster.count > (max_regional_size // zones_per_region):
        return None

    ec2_cost = zones_per_region * cluster.annual_cost

    # Account for the clusters and replication costs
    kafka_costs = {"kafka.zonal-clusters": ec2_cost}

    cluster.cluster_type = "kafka"
    clusters = Clusters(
        annual_costs=kafka_costs,
        zonal=[cluster] * zones_per_region,
        regional=[],
        services=[],
    )

    return CapacityPlan(
        requirements=Requirements(
            zonal=[requirement] * zones_per_region, regrets=regrets
        ),
        candidate_clusters=clusters,
    )


class NflxKafkaArguments(BaseModel):
    copies_per_region: int = Field(
        default=2,
        description=(
            "How many copies of the data will exist e.g. RF=3. If not supplied"
            " this will be determined from mode"
        ),
    )
    cluster_type: ClusterType = Field(
        default=ClusterType.ha,
        description="If the cluster is 'strong' consistency or 'high availability'",
    )
    retention: str = Field(
        default="PT8H", description="How long to retain data in this cluster."
    )
    hot_retention: str = Field(
        default="PT10M",
        description=(
            "How long to retain data in page cache for consumers to skip IO. "
            "Typically consumers lag under 10s, but ensure up to 10M can be handled"
        ),
    )
    require_local_disks: bool = Field(
        default=False,
        description="If local (ephemeral) drives are required",
    )
    require_attached_disks: bool = Field(
        default=False,
        description="If attached (ebs) drives are required",
    )
    required_zonal_size: Optional[int] = Field(
        default=None,
        description="Require zonal clusters to be this size (force vertical scaling)",
    )
    max_regional_size: int = Field(
        default=150,
        description="What is the maximum size of a cluster in this region",
    )
    max_local_disk_gib: int = Field(
        default=5120,
        description=(
            "The maximum amount of data we store per node. Used to limit "
            "recovery duration on failure."
        ),
    )
    min_instance_cpu: int = Field(
        default=2,
        description="The minimum number of instance CPU to allow",
    )
    min_instance_memory_gib: int = Field(
        default=12,
        description="The minimum amount of instance memory to allow",
    )


class NflxKafkaCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        cluster_type: ClusterType = extra_model_arguments.get(
            "cluster_type", ClusterType.ha
        )
        default_replication = 2
        if cluster_type == ClusterType.strong:
            default_replication = 3
        copies_per_region: int = extra_model_arguments.get(
            "copies_per_region", default_replication
        )
        if cluster_type == ClusterType.strong and copies_per_region < 3:
            raise ValueError("Strong consistency and RF<3 doesn't work")

        max_regional_size: int = extra_model_arguments.get("max_regional_size", 150)
        # Very large nodes are hard to cache warm
        max_local_disk_gib: int = extra_model_arguments.get(
            "max_local_disk_gib", 1024 * 5
        )
        min_instance_cpu: int = extra_model_arguments.get("min_instance_cpu", 2)
        min_instance_memory_gib: int = extra_model_arguments.get(
            "min_instance_memory_gib", 12
        )
        hot_retention_seconds: float = iso_to_seconds(
            extra_model_arguments.get("hot_retention", "PT10M")
        )
        require_local_disks: bool = extra_model_arguments.get(
            "require_local_disks", False
        )
        require_attached_disks: bool = extra_model_arguments.get(
            "require_attached_disks", False
        )
        required_zone_size: Optional[int] = extra_model_arguments.get(
            "required_zone_size", None
        )

        return _estimate_kafka_cluster_zonal(
            instance=instance,
            drive=drive,
            desires=desires,
            zones_per_region=context.zones_in_region,
            copies_per_region=copies_per_region,
            require_local_disks=require_local_disks,
            require_attached_disks=require_attached_disks,
            required_zone_size=required_zone_size,
            max_regional_size=max_regional_size,
            max_local_disk_gib=max_local_disk_gib,
            min_instance_cpu=min_instance_cpu,
            min_instance_memory_gib=min_instance_memory_gib,
            hot_retention_seconds=hot_retention_seconds,
        )

    @staticmethod
    def description():
        return "Netflix Streaming Kafka Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxKafkaArguments.model_json_schema()

    @staticmethod
    def allowed_cloud_drives() -> Tuple[Optional[str], ...]:
        # Kafka at Netflix only supports 3rd gen SSD storage
        return ("gp3",)

    @staticmethod
    def default_desires(user_desires, extra_model_arguments: Dict[str, Any]):
        # Default to 10MiB/s and a single reader
        concurrent_readers = max(
            1, int(user_desires.query_pattern.estimated_read_per_second.mid)
        )
        query_pattern = user_desires.query_pattern.model_dump()
        if "estimated_mean_write_size_bytes" in query_pattern:
            write_bytes = Interval(**query_pattern["estimated_mean_write_size_bytes"])
        else:
            write_bytes = certain_int(10 * 1024 * 1024)

        read_bytes = write_bytes.scale(concurrent_readers)
        retention = extra_model_arguments.get("retention", "PT8H")
        retention_secs = iso_to_seconds(retention)

        # write throughput * retention = usage
        state_gib = (write_bytes.mid * retention_secs) / GIB_IN_BYTES

        return CapacityDesires(
            query_pattern=QueryPattern(
                access_pattern=AccessPattern.throughput,
                access_consistency=GlobalConsistency(
                    same_region=Consistency(
                        target_consistency=AccessConsistency.read_your_writes
                    ),
                    cross_region=Consistency(
                        target_consistency=AccessConsistency.never
                    ),
                ),
                estimated_mean_write_size_bytes=write_bytes,
                estimated_mean_read_size_bytes=read_bytes,
                # How much on-cpu time to read or write 1 MiB
                # 42 MiB / second in 3 CPU seconds = 14 MiB / CPU second
                # This = 1000 / 14 = 71 millisecond / MiB
                # To read 1 MiB costs 71 millisecond of CPU time
                # Later tuned down based on training data
                estimated_mean_read_latency_ms=Interval(
                    low=20, mid=40, high=75, confidence=0.98
                ),
                # Wild-Ass-Guess that writes are slightly cheaper
                # Because no disks/context switching??
                estimated_mean_write_latency_ms=Interval(
                    low=20, mid=30, high=75, confidence=0.98
                ),
            ),
            data_shape=DataShape(
                estimated_state_size_gib=Interval(
                    low=state_gib * 0.5,
                    mid=state_gib,
                    high=state_gib * 2,
                    confidence=0.98,
                ),
                # Gandalf, system stuffs
                reserved_instance_app_mem_gib=1,
                # Connection overhead, kernel, etc ...
                reserved_instance_system_mem_gib=3,
            ),
        )


nflx_kafka_capacity_model = NflxKafkaCapacityModel()



from abc import ABC
from abc import abstractmethod


class HeadroomStrategy(ABC):
    @abstractmethod
    def calculate_reserved_headroom(self, effective_cpu: float) -> float:
        pass


class QueuingBasedHeadroomStrategy(HeadroomStrategy):
    """
    The headroom should typically indicate the percentage of CPU that should be
    reserved to ensure sensible performance profile.

    This calculates headroom using the Erlang-C staffing formula with P_Q=30:

    See /notebooks/headroom-estimator.ipynb for details
    """

    def calculate_reserved_headroom(self, effective_cpu: float) -> float:
        return 0.712 / (effective_cpu**0.448)



import logging
import math
from enum import Enum
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import compute_stateful_zone
from service_capacity_modeling.models.common import network_services
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import simple_network_mbps
from service_capacity_modeling.models.common import sqrt_staffed_cores
from service_capacity_modeling.models.common import working_set_from_drive_and_slo
from service_capacity_modeling.models.utils import next_n
from service_capacity_modeling.stats import dist_for_interval

logger = logging.getLogger(__name__)


class Replication(str, Enum):
    none = "none"
    sets = "sets"
    evicts = "evicts"


def calculate_read_cpu_time_evcache_ms(read_size_bytes: float) -> float:
    # Fitted a curve based on some data that we crunched from couple of
    # read heavy clusters
    # In memory
    #  250 bit - 10 micros
    # 1520 bit - 41 micros
    # 8250 bit - 66 micros
    # On disk
    # 24   KiB - 133 micros
    # 40   KiB - 158 top of our curve
    # Fit a logistic curve, requiring it to go through first
    # point
    read_latency_ms = 979.4009 + (-0.06853492 - 979.4009) / math.pow(
        (1 + math.pow(read_size_bytes / 13061.23, 0.180864)), 0.0002819491
    )
    return max(read_latency_ms, 0.005)


def calculate_spread_cost(cluster_size: int, max_cost=100000, min_cost=0.0) -> float:
    if cluster_size > 10:
        return min_cost
    if cluster_size < 2:
        return max_cost
    return min_cost + (max_cost - cluster_size * (max_cost - min_cost) / 30.0)


def _estimate_evcache_requirement(
    instance: Instance,
    desires: CapacityDesires,
    working_set: Optional[float],
    copies_per_region: int,
) -> Tuple[CapacityRequirement, Tuple[str, ...]]:
    """Estimate the capacity required for one zone given a regional desire

    The input desires should be the **regional** desire, and this function will
    return the zonal capacity requirement
    """
    needed_cores = normalize_cores(
        core_count=sqrt_staffed_cores(desires),
        target_shape=instance,
        reference_shape=desires.reference_shape,
    )

    # For tier 0, we double the number of cores to account for caution
    if desires.service_tier == 0:
        needed_cores = needed_cores * 2

    # (Arun): Keep 20% of available bandwidth for cache warmer
    needed_network_mbps = simple_network_mbps(desires) * 1.25

    needed_disk = math.ceil(
        desires.data_shape.estimated_state_size_gib.mid,
    )

    regrets: Tuple[str, ...] = ("spend", "mem")
    state_size = desires.data_shape.estimated_state_size_gib
    item_count = desires.data_shape.estimated_state_item_count
    payload_greater_than_classic = False
    if state_size is not None and item_count is not None and item_count.mid != 0:
        payload_size = (state_size.mid * 1024.0 * 1024.0 * 1024.0) / (item_count.mid)
        if payload_size > 200.0:
            payload_greater_than_classic = True
    else:
        if desires.query_pattern.estimated_mean_read_size_bytes.mid > 200.0:
            payload_greater_than_classic = True

    # (Arun): As of 2021 we are using ephemerals exclusively and do not
    # use cloud drives
    if working_set is None or (
        desires.data_shape.estimated_state_size_gib.mid < 110.0
        and payload_greater_than_classic
    ):
        # We can't currently store data on cloud drives, but we can put the
        # dataset into memory!
        needed_memory = float(needed_disk)
        needed_disk = 0
    else:
        # We can store data on fast ephems (reducing the working set that must
        # be kept in RAM)
        needed_memory = float(working_set) * float(needed_disk)
        regrets = ("spend", "disk", "mem")

    # For EVCache, writes go to all zones
    # Regional reads can also go to any one zone due to app's zone affinity
    needed_cores = max(1, needed_cores)
    logger.debug(
        "Need (cpu, mem, disk, working) = (%s, %s, %s, %f)",
        needed_cores,
        needed_memory,
        needed_disk,
        working_set,
    )

    return (
        CapacityRequirement(
            requirement_type="evcache-zonal",
            reference_shape=desires.reference_shape,
            cpu_cores=certain_int(needed_cores),
            mem_gib=certain_float(needed_memory),
            disk_gib=certain_float(needed_disk),
            network_mbps=certain_float(needed_network_mbps),
            context={
                "working_set": working_set,
                "replication_factor": copies_per_region,
            },
        ),
        regrets,
    )


def _upsert_params(cluster, params):
    if cluster.cluster_params:
        cluster.cluster_params.update(params)
    else:
        cluster.cluster_params = params


# pylint: disable=too-many-locals
def _estimate_evcache_cluster_zonal(  # noqa: C901,E501 pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    desires: CapacityDesires,
    context: RegionContext,
    copies_per_region: int = 3,
    max_local_disk_gib: int = 2048,
    max_regional_size: int = 10000,
    min_instance_memory_gib: int = 12,
    cross_region_replication: Replication = Replication.none,
) -> Optional[CapacityPlan]:

    # EVCache doesn't like to deploy on single CPU instances
    if instance.cpu < 2:
        return None

    # EVCache doesn't like to deploy to instances with < 7 GiB of ram
    if instance.ram_gib < min_instance_memory_gib:
        return None

    if instance.drive is None:
        return None

    # Based on the disk latency and the read latency SLOs we adjust our
    # working set to keep more or less data in RAM. Faster drives need
    # less fronting RAM.
    ws_drive = instance.drive or drive

    if ws_drive:
        working_set = working_set_from_drive_and_slo(
            drive_read_latency_dist=dist_for_interval(ws_drive.read_io_latency_ms),
            read_slo_latency_dist=dist_for_interval(
                desires.query_pattern.read_latency_slo_ms
            ),
            estimated_working_set=desires.data_shape.estimated_working_set_percent,
            # Caches have very tight latency SLOs, so we target a high
            # percentile of the drive latency distribution for WS calculation
            target_percentile=0.99,
        ).mid
    else:
        working_set = None

    requirement, regrets = _estimate_evcache_requirement(
        instance=instance,
        desires=desires,
        working_set=working_set,
        copies_per_region=copies_per_region,
    )

    # Account for sidecars and base system memory
    base_mem = (
        desires.data_shape.reserved_instance_app_mem_gib
        + desires.data_shape.reserved_instance_system_mem_gib
    )

    # (Arun) We currently reserve extra memory for the OS as instances get
    # larger to account for additional overhead. Note that the
    # reserved_instance_system_mem_gib has a base of 1 GiB OSMEM so this
    # just represents the variable component
    def reserve_memory(instance_mem_gib):
        # (Joey) From the chart it appears to be about a 3% overhead for
        # OS memory.
        variable_os = int(instance_mem_gib * 0.03)
        return base_mem + variable_os

    requirement.context["osmem"] = reserve_memory(instance.ram_gib)
    # EVCache clusters aim to be at least 2 nodes per zone to start
    # out with for tier 0
    min_count = 0
    if desires.service_tier < 1:
        min_count = 2

    is_disk_io_constraint: bool = requirement.disk_gib.mid > 0.0
    adjusted_disk_io_needed = 0.0
    read_write_ratio = 0.0
    if is_disk_io_constraint:
        reads_per_sec = desires.query_pattern.estimated_read_per_second.mid
        writes_per_sec = desires.query_pattern.estimated_write_per_second.mid
        read_size = desires.query_pattern.estimated_mean_read_size_bytes.mid
        write_size = desires.query_pattern.estimated_mean_write_size_bytes.mid
        read_disk_io_needed = reads_per_sec * read_size
        write_disk_io_needed = writes_per_sec * write_size
        adjusted_disk_io_needed = read_disk_io_needed + write_disk_io_needed
        # Giving headroom for cachewarming and region squeeze
        adjusted_disk_io_needed = 1.4 * adjusted_disk_io_needed
        read_write_ratio = reads_per_sec / (reads_per_sec + writes_per_sec)

    cluster = compute_stateful_zone(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=int(requirement.disk_gib.mid),
        needed_memory_gib=int(requirement.mem_gib.mid),
        needed_network_mbps=requirement.network_mbps.mid,
        # EVCache doesn't use cloud drives to store data, we will have
        # accounted for the data going on drives or memory via working set
        max_local_disk_gib=max_local_disk_gib,
        # EVCache clusters should be balanced per zone
        cluster_size=lambda x: next_n(x, copies_per_region),
        min_count=max(min_count, 0),
        # Sidecars and Variable OS Memory
        reserve_memory=lambda x: base_mem,
        adjusted_disk_io_needed=adjusted_disk_io_needed,
        read_write_ratio=read_write_ratio,
    )
    # Communicate to the actual provision that if we want reduced RF
    params = {"evcache.copies": copies_per_region}
    _upsert_params(cluster, params)

    # evcache clusters generally should try to stay under some total number
    # of nodes. Orgs do this for all kinds of reasons such as
    #   * Security group limits. Since you must have < 500 rules if you're
    #       ingressing public ips)
    #   * Maintenance. If your restart script does one node at a time you want
    #       smaller clusters so your restarts don't take months.
    #   * NxN network issues. Sometimes smaller clusters of bigger nodes
    #       are better for network propagation
    if cluster.count > (max_regional_size // copies_per_region):
        return None

    services = []
    if cross_region_replication is Replication.sets:
        services.extend(
            network_services("evcache", context, desires, copies_per_region)
        )
    elif cross_region_replication is Replication.evicts:
        modified = desires.model_copy(deep=True)
        # Assume that DELETES replicating cross region mean 128 bytes
        # of key per evict.
        modified.query_pattern.estimated_mean_write_size_bytes = certain_int(128)
        services.extend(
            network_services("evcache", context, modified, copies_per_region)
        )

    ec2_cost = copies_per_region * cluster.annual_cost
    spread_cost = calculate_spread_cost(cluster.count)

    # Account for the clusters and replication costs
    evcache_costs = {
        "evcache.zonal-clusters": ec2_cost,
        "evcache.spread.cost": spread_cost,
    }

    for s in services:
        evcache_costs[f"{s.service_type}"] = s.annual_cost

    cluster.cluster_type = "evcache"
    clusters = Clusters(
        annual_costs=evcache_costs,
        zonal=[cluster] * copies_per_region,
        regional=[],
        services=services,
    )

    return CapacityPlan(
        requirements=Requirements(
            zonal=[requirement] * copies_per_region, regrets=regrets
        ),
        candidate_clusters=clusters,
    )


class NflxEVCacheArguments(BaseModel):
    copies_per_region: int = Field(
        default=3,
        description="How many copies of the data will exist e.g. RF=3. If not supplied"
        " this will be deduced from tier",
    )
    max_regional_size: int = Field(
        default=10000,
        description="What is the maximum size of a cluster in this region",
    )
    max_local_disk_gib: int = Field(
        default=6144,
        description="The maximum amount of data we store per machine",
    )
    min_instance_memory_gib: int = Field(
        default=12,
        description="The minimum amount of instance memory to allow",
    )
    cross_region_replication: Replication = Field(
        default=Replication.none,
        description=(
            "Whether this evcache service does cross region replication. "
            "By default we do no replication"
        ),
    )


class NflxEVCacheCapacityModel(CapacityModel):
    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # (Arun) EVCache defaults to RF=3 for tier 0 and tier 1
        default_copies = context.zones_in_region
        copies_per_region: int = extra_model_arguments.get(
            "copies_per_region", default_copies
        )
        max_regional_size: int = extra_model_arguments.get("max_regional_size", 10000)
        # Very large nodes are hard to cache warm
        max_local_disk_gib: int = extra_model_arguments.get(
            "max_local_disk_gib", 1024 * 6
        )
        # Very small nodes are hard to run memcache on
        # (Arun) We do not deploy to less than 12 GiB
        min_instance_memory_gib: int = extra_model_arguments.get(
            "min_instance_memory_gib", 12
        )
        cross_region_replication = Replication(
            extra_model_arguments.get("cross_region_replication", "none")
        )

        return _estimate_evcache_cluster_zonal(
            instance=instance,
            drive=drive,
            desires=desires,
            copies_per_region=copies_per_region,
            max_regional_size=max_regional_size,
            max_local_disk_gib=max_local_disk_gib,
            min_instance_memory_gib=min_instance_memory_gib,
            cross_region_replication=cross_region_replication,
            context=context,
        )

    @staticmethod
    def description():
        return "Netflix Streaming EVCache (memcached) Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxEVCacheArguments.model_json_schema()

    @staticmethod
    def default_desires(user_desires, extra_model_arguments: Dict[str, Any]):
        acceptable_consistency = {
            AccessConsistency.best_effort,
            AccessConsistency.never,
            None,
        }

        for key, value in user_desires.query_pattern.access_consistency:
            if value.target_consistency not in acceptable_consistency:
                raise ValueError(
                    f"EVCache can only provide {acceptable_consistency} access."
                    f"User asked for {key}={value}"
                )

        estimated_read_size: Interval = Interval(
            **user_desires.query_pattern.model_dump().get(
                "estimated_mean_read_size_bytes",
                user_desires.query_pattern.model_dump().get(
                    "estimated_mean_write_size_bytes",
                    {"low": 16, "mid": 1024, "high": 65536, "confidence": 0.95},
                ),
            )
        )
        estimated_read_latency_ms: Interval = Interval(
            low=calculate_read_cpu_time_evcache_ms(estimated_read_size.low),
            mid=calculate_read_cpu_time_evcache_ms(estimated_read_size.mid),
            high=calculate_read_cpu_time_evcache_ms(estimated_read_size.high),
            confidence=estimated_read_size.confidence,
        )

        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.best_effort
                        ),
                        # By default EVCache does not globally replicate
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.never
                        ),
                    ),
                    estimated_mean_read_size_bytes=estimated_read_size,
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=512, high=1024, confidence=0.95
                    ),
                    # evcache read latency is sensitive to payload size
                    # so this is computed above
                    estimated_mean_read_latency_ms=estimated_read_latency_ms,
                    # evcache bulk puts usually take slightly longer
                    estimated_mean_write_latency_ms=Interval(
                        low=0.01, mid=0.01, high=0.01, confidence=0.98
                    ),
                    # Assume point queries, "1 millisecond SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.1,
                        maximum_value=4,
                        low=0.15,
                        mid=0.5,
                        high=1,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=1,
                        high=5,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    # (Arun): Most latency sensitive < 600GiB
                    estimated_state_size_gib=Interval(
                        low=10, mid=100, high=600, confidence=0.98
                    ),
                    # (Arun): The management sidecar takes 512 MiB
                    reserved_instance_app_mem_gib=1,
                    # account for the memcached connection memory
                    # and system requirements.
                    # (Arun) We currently use 1 GiB for connection memory
                    reserved_instance_system_mem_gib=(1 + 2),
                ),
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.throughput,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.best_effort
                        ),
                        # By default EVCache does not globally replicate
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.never
                        ),
                    ),
                    estimated_mean_read_size_bytes=estimated_read_size,
                    estimated_mean_write_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    # evcache read latency is sensitive to payload size
                    # so this is computed above
                    estimated_mean_read_latency_ms=estimated_read_latency_ms,
                    # evcache bulk puts usually take slightly longer
                    estimated_mean_write_latency_ms=Interval(
                        low=0.01, mid=0.01, high=0.01, confidence=0.98
                    ),
                    # Assume they're multi-getting -> slow reads
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.1,
                        maximum_value=10,
                        low=0.2,
                        mid=0.5,
                        high=1,
                        confidence=0.98,
                    ),
                    # Assume they're multi-setting writes
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.1,
                        maximum_value=10,
                        low=0.2,
                        mid=0.5,
                        high=1,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    # Typical ML models go up to TiB
                    estimated_state_size_gib=Interval(
                        low=10, mid=100, high=1000, confidence=0.98
                    ),
                    # (Arun): The management sidecar takes 512 MiB
                    reserved_instance_app_mem_gib=1,
                    # account for the memcached connection memory
                    # and system requirements.
                    # (Arun) We currently use 1 GiB base for connection memory
                    reserved_instance_system_mem_gib=(1 + 2),
                ),
            )


nflx_evcache_capacity_model = NflxEVCacheCapacityModel()



import argparse
import json
import os

import boto3


def extract_3yr_upfront_price(price_data):
    instance_type = price_data["product"]["attributes"]["instanceType"]

    # Look through Reserved terms
    if price_data.get("terms") is None:
        return None

    reserved_terms = price_data["terms"].get("Reserved", {})

    for term in reserved_terms.values():
        term_attrs = term["termAttributes"]
        # Find 3yr All Upfront terms
        if (
            term_attrs.get("LeaseContractLength") == "3yr"
            and term_attrs.get("PurchaseOption") == "All Upfront"
            and term_attrs.get("OfferingClass") == "standard"
        ):

            # Get upfront fee
            for dim in term["priceDimensions"].values():
                if dim["unit"] == "Quantity":
                    upfront = float(dim["pricePerUnit"]["USD"])
                    annual_cost = upfront / 3
                    annual_cost_rounded = round(annual_cost, 2)
                    print(f"{instance_type}: {annual_cost_rounded}")
                    return annual_cost_rounded

    return None


def fetch_pricing(region: str):
    # Initialize pricing client
    pricing_client = boto3.client("pricing", region_name=region)

    # Get all EC2 pricing for Reserved Instances
    paginator = pricing_client.get_paginator("get_products")

    instances = {}

    filter_params = {
        "ServiceCode": "AmazonEC2",
        "Filters": [
            {
                "Type": "TERM_MATCH",
                "Field": "location",
                "Value": "US East (N. Virginia)",
            },
            {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"},
            {"Type": "TERM_MATCH", "Field": "tenancy", "Value": "Shared"},
            {"Type": "TERM_MATCH", "Field": "operatingSystem", "Value": "Linux"},
            {"Type": "TERM_MATCH", "Field": "preInstalledSw", "Value": "NA"},
        ],
    }

    for page in paginator.paginate(**filter_params):
        for price_item in page["PriceList"]:
            price_data = json.loads(price_item)

            # Extract instance type
            attributes = price_data.get("product", {}).get("attributes", {})
            instance_type = attributes.get("instanceType")

            if not instance_type:
                continue

            annual_cost = extract_3yr_upfront_price(price_data)
            if annual_cost:
                instances[instance_type] = {"annual_cost": annual_cost}
                if "deprecated" in instance_type.lower():
                    instances[instance_type]["lifecycle"] = "deprecated"

    # Create final output structure
    # we bolt on the other info, as a hack until we can improve prior layers
    output = {
        "us-east-1": {
            "instances": instances,
            "drives": {},
            "services": {},
            "zones_in_region": 3,
        }
    }

    # Write to JSON file
    project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    output_file = os.path.join(
        project_root,
        "service_capacity_modeling",
        "hardware",
        "profiles",
        "pricing",
        "aws",
        "3yr-reserved_ec2.json",
    )
    with open(output_file, "w", encoding="utf-8") as f:
        json.dump(output, f, indent=2, sort_keys=True)
        f.write("\n")

    print(f"Pricing data written to {output_file}")


def main():
    parser = argparse.ArgumentParser(
        description="Fetch EC2 Reserved Instance pricing data."
    )
    parser.add_argument(
        "--region",
        type=str,
        default="us-east-1",
        help="AWS region (default: us-east-1)",
    )
    args = parser.parse_args()

    fetch_pricing(args.region)


if __name__ == "__main__":
    main()



import math
from typing import Iterable
from typing import List
from typing import Set
from typing import Tuple

from service_capacity_modeling.models import CapacityPlan


def reduce_by_family(plans: Iterable[CapacityPlan]) -> List[CapacityPlan]:
    """Groups a potential set of clusters by hardware family sorted by cost.

    Useful for showing different family options.
    """
    zonal_families: Set[Tuple[Tuple[str, str], ...]] = set()
    regional_families: Set[Tuple[Tuple[str, str], ...]] = set()

    result: List[CapacityPlan] = []
    for plan in plans:
        topo = plan.candidate_clusters
        regional_type: Tuple[Tuple[str, str], ...] = tuple()
        zonal_type: Tuple[Tuple[str, str], ...] = tuple()

        if topo.regional:
            regional_type = tuple(
                sorted({(c.cluster_type, c.instance.family) for c in topo.regional})
            )

        if topo.zonal:
            zonal_type = tuple(
                sorted({(c.cluster_type, c.instance.family) for c in topo.zonal})
            )

        if not (zonal_type in zonal_families and regional_type in regional_families):
            result.append(plan)

        regional_families.add(regional_type)
        zonal_families.add(zonal_type)

    return result


# https://stackoverflow.com/questions/14267555/find-the-smallest-power-of-2-greater-than-or-equal-to-n-in-python
def next_power_of_2(y: float) -> int:
    x = int(y)
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


def next_n(x: float, n: float) -> int:
    return int(math.ceil(x / n)) * int(n)



import math
from datetime import timedelta
from decimal import Decimal

import isodate
from isodate import parse_duration

DURATION_1M = 2592000
DURATION_1Y = 31536000
D_YEAR = Decimal(DURATION_1Y)
D_MONTH = Decimal(DURATION_1M)


def _iso_to_proto_duration(iso_duration: str):
    parsed = parse_duration(iso_duration)
    return f"{int(parsed.total_seconds())}s"


def iso_to_seconds(iso_duration: str, unlimited=0) -> int:
    if iso_duration == "unlimited":
        return unlimited
    parsed = isodate.parse_duration(iso_duration)
    # isodate package parse_duration returns either timedelta OR Duration
    if isinstance(parsed, isodate.Duration):
        return math.ceil(
            parsed.years * D_YEAR
            + parsed.months * D_MONTH
            + Decimal(parsed.tdelta.total_seconds())
        )
    return int(parsed.total_seconds())


def _iso_to_timedelta(iso_duration: str, unlimited=0) -> timedelta:
    return timedelta(seconds=iso_to_seconds(iso_duration, unlimited))
